# pylint: disable=missing-class-docstring,no-self-use
from __future__ import annotations

import sys
import unittest

import claripy
from claripy import FPS, fpToSBV, fpToUBV
from claripy.fp import RM


class TestZ3(unittest.TestCase):
    """
    A class used for testing z3
    """

    def test_extrema(self):
        """
        Test the _extrema function within the z3 backend
        """
        z = claripy.backends.z3

        s = z.solver()
        x = claripy.BVS("x", 8)
        range_ = (0, 2**x.length - 1)

        assert z.min(x, solver=s) == range_[0]
        assert z.max(x, solver=s) == range_[1]

        for i in range(0, range_[1] + 1, 10):
            # ==
            assert z.min(x, solver=s, extra_constraints=(x == i,)) == i
            assert z.max(x, solver=s, extra_constraints=(x == i,)) == i
            # <=
            assert z.min(x, solver=s, extra_constraints=(x <= i,)) == range_[0]
            assert z.max(x, solver=s, extra_constraints=(x <= i,)) == i
            # >=
            assert z.min(x, solver=s, extra_constraints=(x >= i,)) == i
            assert z.max(x, solver=s, extra_constraints=(x >= i,)) == range_[1]

    def test_str2int(self):
        """
        Test the str_to_int_unlimited function.
        """

        s2i = claripy.backends.backend_z3.str_to_int_unlimited
        CHUNK_SIZE = sys.get_int_max_str_digits() if hasattr(sys, "get_int_max_str_digits") else 640

        assert s2i("0") == 0
        assert s2i("1337") == 1337
        assert s2i("1337133713371337") == 1337133713371337
        assert s2i("1" + "0" * 639) == 10**639
        assert s2i("1" + "0" * 640) == 10**640
        assert s2i("1" + "0" * 641) == 10**641
        assert s2i("1" + "0" * 640 + "1") == 10**641 + 1
        assert s2i("1" + "0" * 8192) == 10**8192

        assert s2i("1" + "0" * (CHUNK_SIZE - 1)) == 10 ** (CHUNK_SIZE - 1)
        assert s2i("1" + "0" * CHUNK_SIZE) == 10**CHUNK_SIZE
        assert s2i("1" + "0" * (CHUNK_SIZE + 1)) == 10 ** (CHUNK_SIZE + 1)

        assert s2i("-0") == 0
        assert s2i("-1") == -1
        assert s2i("-1" + "0" * CHUNK_SIZE) == -(10**CHUNK_SIZE)

    def test_int2str(self):
        """
        Test the int_to_str_unlimited function.
        """

        i2s = claripy.backends.backend_z3.int_to_str_unlimited
        CHUNK_SIZE = sys.get_int_max_str_digits() if hasattr(sys, "get_int_max_str_digits") else 640

        assert i2s(0) == "0"
        assert i2s(-1) == "-1"
        assert i2s(1337) == "1337"
        assert i2s(10**8192) == "1" + "0" * 8192
        assert i2s(10**CHUNK_SIZE) == "1" + "0" * CHUNK_SIZE

    def test_get_long_strings(self):
        # Python 3.11 introduced limits in decimal-to-int conversion. we bypass it by using the str_to_int_unlimited
        # method.
        z3 = claripy.backends.backend_z3

        s = claripy.backends.z3.solver()
        backend = z3.BackendZ3()
        x = claripy.BVS("x", 16385 * 8)
        backend.add(s, [x == 10**16384])
        d = backend.eval(x, 1, solver=s)
        assert d == [10**16384]

    def test_abstract_sbv(self):
        """
        Test abstracting fpToSBV
        """

        f = FPS("f", claripy.FSORT_FLOAT)
        sbv = fpToSBV(RM.RM_NearestTiesEven, f, 32)

        sbv2 = claripy.backends.z3.simplify(sbv)
        assert sbv2.length == 32

    def test_abstract_ubv(self):
        """
        Test abstracting fpToUBV
        """

        f = FPS("f", claripy.FSORT_FLOAT)
        sbv = fpToUBV(RM.RM_NearestTiesEven, f, 32)

        sbv2 = claripy.backends.z3.simplify(sbv)
        assert sbv2.length == 32

    def test_abstract_rotate_left(self):
        """
        Test abstracting rotate_left
        """

        x = claripy.BVS("x", 32)
        y = claripy.BVS("y", 32)
        r = claripy.RotateLeft(x, y)

        r2 = claripy.backends.z3.simplify(r)
        assert r2.length == 32

    def test_abstract_rotate_right(self):
        """
        Test abstracting rotate_right
        """

        x = claripy.BVS("x", 32)
        y = claripy.BVS("y", 32)
        r = claripy.RotateRight(x, y)

        r2 = claripy.backends.z3.simplify(r)
        assert r2.length == 32


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from claripy import backends, frontend


class Solver(
    frontend.mixin.ConcreteHandlerMixin,
    frontend.mixin.EagerResolutionMixin,
    frontend.mixin.ConstraintFilterMixin,
    frontend.mixin.ConstraintDeduplicatorMixin,
    frontend.mixin.SimplifySkipperMixin,
    frontend.mixin.SatCacheMixin,
    frontend.mixin.ModelCacheMixin,
    frontend.mixin.ConstraintExpansionMixin,
    frontend.mixin.SimplifyHelperMixin,
    frontend.FullFrontend,
):
    """Solver is the default Claripy frontend. It uses Z3 as the backend solver by default."""

    def __init__(self, backend=backends.z3, **kwargs):
        super().__init__(backend, **kwargs)


class SolverCacheless(
    frontend.mixin.ConcreteHandlerMixin,
    frontend.mixin.EagerResolutionMixin,
    frontend.mixin.ConstraintFilterMixin,
    frontend.mixin.ConstraintDeduplicatorMixin,
    frontend.mixin.SimplifySkipperMixin,
    frontend.FullFrontend,
):
    """SolverCacheless is a Solver without caching. It uses Z3 as the backend solver by default."""

    def __init__(self, backend=backends.z3, **kwargs):
        super().__init__(backend, **kwargs)


class SolverReplacement(
    frontend.mixin.ConcreteHandlerMixin,
    frontend.mixin.ConstraintDeduplicatorMixin,
    frontend.ReplacementFrontend,
):
    """SolverReplacement is a frontend wrapper that replaces constraints with their solutions."""

    def __init__(self, actual_frontend=None, **kwargs):
        actual_frontend = Solver() if actual_frontend is None else actual_frontend
        super().__init__(actual_frontend, **kwargs)


class SolverHybrid(
    frontend.mixin.ConcreteHandlerMixin,
    frontend.mixin.EagerResolutionMixin,
    frontend.mixin.ConstraintFilterMixin,
    frontend.mixin.ConstraintDeduplicatorMixin,
    frontend.mixin.SimplifySkipperMixin,
    # TODO: frontend.mixin.ConstraintExpansionMixin,
    frontend.HybridFrontend,
):
    """SolverHybrid is a frontend that uses an exact solver and an approximate solver."""

    def __init__(  # pylint:disable=too-many-positional-arguments
        self,
        exact_frontend=None,
        approximate_frontend=None,
        complex_auto_replace=True,
        replace_constraints=True,
        track=False,
        approximate_first=False,
        **kwargs,
    ):
        exact_frontend = Solver(track=track) if exact_frontend is None else exact_frontend
        approximate_frontend = (
            SolverReplacement(
                actual_frontend=SolverVSA(),
                complex_auto_replace=complex_auto_replace,
                replace_constraints=replace_constraints,
            )
            if approximate_frontend is None
            else approximate_frontend
        )
        super().__init__(exact_frontend, approximate_frontend, approximate_first=approximate_first, **kwargs)


class SolverVSA(
    frontend.mixin.ConcreteHandlerMixin,
    frontend.mixin.ConstraintFilterMixin,
    frontend.LightFrontend,
):
    """SolverVSA is a thin frontend to the VSA backend solver."""

    def __init__(self, **kwargs):
        super().__init__(backends.vsa, **kwargs)


class SolverConcrete(
    frontend.mixin.ConcreteHandlerMixin,
    frontend.mixin.ConstraintFilterMixin,
    frontend.LightFrontend,
):
    """SolverConcrete is a thin frontend to the Concrete backend solver."""

    def __init__(self, **kwargs):
        super().__init__(backends.concrete, **kwargs)


class SolverStrings(
    # TODO: Figure ot if we need to use all these frontend.mixins
    frontend.mixin.ConcreteHandlerMixin,
    frontend.mixin.ConstraintFilterMixin,
    frontend.mixin.ConstraintDeduplicatorMixin,
    frontend.mixin.EagerResolutionMixin,
    frontend.FullFrontend,
):
    """SolverStrings is a frontend that uses Z3 to solve string constraints."""

    def __init__(self, *args, backend=backends.z3, **kwargs):
        super().__init__(backend, *args, **kwargs)


#
# Composite solving
#


class SolverCompositeChild(
    frontend.mixin.ConstraintDeduplicatorMixin,
    frontend.mixin.SatCacheMixin,
    frontend.mixin.SimplifySkipperMixin,
    frontend.mixin.ModelCacheMixin,
    frontend.FullFrontend,
):
    """SolverCompositeChild is a frontend that is used as a child in a SolverComposite."""

    def __init__(self, backend=backends.z3, **kwargs):
        super().__init__(backend, **kwargs)

    def __repr__(self):
        return f"<SolverCompositeChild with {len(self.variables)} variables>"


class SolverComposite(
    frontend.mixin.ConcreteHandlerMixin,
    frontend.mixin.EagerResolutionMixin,
    frontend.mixin.ConstraintFilterMixin,
    frontend.mixin.ConstraintDeduplicatorMixin,
    frontend.mixin.SatCacheMixin,
    frontend.mixin.SimplifySkipperMixin,
    frontend.mixin.SimplifyHelperMixin,
    frontend.mixin.ConstraintExpansionMixin,
    frontend.mixin.CompositedCacheMixin,
    frontend.CompositeFrontend,
):
    """SolverComposite is a frontend that composes multiple templated frontends."""

    def __init__(self, template_solver=None, track=False, **kwargs):
        template_solver = SolverCompositeChild(track=track) if template_solver is None else template_solver
        super().__init__(template_solver, track=track, **kwargs)

    def __repr__(self):
        return f"<SolverComposite {id(self):#x}, {len(self._solver_list)} children>"



from __future__ import annotations

from typing import TYPE_CHECKING

from claripy import And, false, is_false
from claripy.errors import UnsatError

if TYPE_CHECKING:
    from collections.abc import Iterable

    from claripy.ast.bool import Bool


class SatCacheMixin:
    """SatCacheMixin is a mixin that caches the satisfiability of the constraints."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._cached_satness = None
        self._cached_unsat_core: tuple[Bool, ...] | None = None

    def _blank_copy(self, c):
        super()._blank_copy(c)
        c._cached_satness = None
        c._cached_unsat_core = None

    def _copy(self, c):
        super()._copy(c)
        c._cached_satness = self._cached_satness
        c._cached_unsat_core = self._cached_unsat_core

    def __getstate__(self):
        return self._cached_satness, self._cached_unsat_core, super().__getstate__()

    def __setstate__(self, s):
        self._cached_satness, self._cached_unsat_core, base_state = s
        super().__setstate__(base_state)

    #
    # SAT caching
    #

    def _add(self, constraints, invalidate_cache=True):
        added = super()._add(constraints, invalidate_cache=invalidate_cache)

        cached_satness = None
        if len(added) > 0:
            if any(c is false() for c in added):
                cached_satness = False
            elif len(added) == 1 and len(self.constraints) < 5:
                added_ = added[0]
                for con in self.constraints:
                    if is_false(And(con.clear_annotations(), added_.clear_annotations())):
                        cached_satness = False
                        self._cached_unsat_core = (con, added)
                        break

        if cached_satness is False:
            self._cached_satness = False
        elif self._cached_satness is True:
            self._cached_satness = None

        return added

    def simplify(self):
        new_constraints = super().simplify()
        if len(new_constraints) > 0 and any(c is false() for c in new_constraints):
            self._cached_satness = False
        return new_constraints

    def satisfiable(self, extra_constraints=(), exact=None):
        if self._cached_satness is False:
            return False
        if self._cached_satness is True and len(extra_constraints) == 0:
            return True
        r = super().satisfiable(extra_constraints=extra_constraints, exact=exact)
        if len(extra_constraints) == 0:
            self._cached_satness = r
        return r

    def check_satisfiability(self, extra_constraints=(), exact=None) -> str:
        if self._cached_satness is False:
            return "UNSAT"
        if self._cached_satness is True and not extra_constraints:
            return "SAT"
        return super().check_satisfiability(extra_constraints=extra_constraints, exact=exact)

    def eval(self, e, n, extra_constraints=(), exact=None):
        if self._cached_satness is False:
            raise UnsatError("cached unsat")
        try:
            r = super().eval(e, n, extra_constraints=extra_constraints, exact=exact)
            self._cached_satness = True
            return r
        except UnsatError:
            if len(extra_constraints) == 0:
                self._cached_satness = False
            raise

    def batch_eval(self, e, n, extra_constraints=(), exact=None):
        if self._cached_satness is False:
            raise UnsatError("cached unsat")
        try:
            r = super().batch_eval(e, n, extra_constraints=extra_constraints, exact=exact)
            self._cached_satness = True
            return r
        except UnsatError:
            if len(extra_constraints) == 0:
                self._cached_satness = False
            raise

    def max(self, e, extra_constraints=(), signed=False, exact=None):
        if self._cached_satness is False:
            raise UnsatError("cached unsat")
        try:
            r = super().max(e, extra_constraints=extra_constraints, signed=signed, exact=exact)
            self._cached_satness = True
            return r
        except UnsatError:
            if len(extra_constraints) == 0:
                self._cached_satness = False
            raise

    def min(self, e, extra_constraints=(), signed=False, exact=None):
        if self._cached_satness is False:
            raise UnsatError("cached unsat")
        try:
            r = super().min(e, extra_constraints=extra_constraints, signed=signed, exact=exact)
            self._cached_satness = True
            return r
        except UnsatError:
            if len(extra_constraints) == 0:
                self._cached_satness = False
            raise

    def solution(self, e, v, extra_constraints=(), exact=None):
        if self._cached_satness is False:
            raise UnsatError("cached unsat")
        try:
            r = super().solution(e, v, extra_constraints=extra_constraints, exact=exact)
            self._cached_satness = True
            return r
        except UnsatError:
            if len(extra_constraints) == 0:
                self._cached_satness = False
            raise

    def unsat_core(self, extra_constraints: tuple[Bool, ...] = ()) -> Iterable[Bool]:
        if self._cached_unsat_core is not None:
            return self._cached_unsat_core
        return super().unsat_core(extra_constraints)



# pylint: disable=missing-class-docstring,no-self-use
from __future__ import annotations

import unittest

from claripy.backends.backend_vsa import StridedInterval


def check_si_fields(si, stride, lb, ub):
    lb &= si.max_int(si.bits)
    ub &= si.max_int(si.bits)
    if si.stride != stride:
        return False
    if si.lower_bound != lb:
        return False
    return si.upper_bound == ub


class TestDivision(unittest.TestCase):
    # non-overlapping
    def test_simple_1(self):
        op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=3)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=2, upper_bound=2)
        # <4>0x1[0x0, 0x1]
        assert check_si_fields(op1.sdiv(op2), 1, 0, 1)
        assert check_si_fields(op1.udiv(op2), 1, 0, 1)

    def test_simple_2(self):
        op1 = StridedInterval(bits=4, stride=1, lower_bound=-6, upper_bound=-4)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=1)
        # <4>0x1[0xa, 0xc]
        assert check_si_fields(op1.sdiv(op2), 1, 10, 12)
        assert check_si_fields(op1.udiv(op2), 1, 10, 12)

    def test_simple_3(self):
        op1 = StridedInterval(bits=4, stride=1, lower_bound=3, upper_bound=-2)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=-2, upper_bound=-2)
        # udiv:  <4>0x1[0x0, 0x1]
        # sdiv:  <4>0x1[0xc, 0x4]
        assert check_si_fields(op1.udiv(op2), 1, 0, 1)
        assert check_si_fields(op1.sdiv(op2), 1, 12, 4)

    def test_simple_4(self):
        # simple case 4 : Result should be zero
        op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=3)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=0, upper_bound=0)
        # BOT
        assert op1.sdiv(op2).is_empty
        assert op1.udiv(op2).is_empty

    def test_both_0_hemisphere(self):
        # Both in 0-hemisphere
        op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=3)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=4, upper_bound=6)
        # udiv: <4>0x0[0x0, 0x0]
        # sdiv: <4>0x0[0x0, 0x0]
        assert check_si_fields(op1.udiv(op2), 0, 0, 0)
        assert check_si_fields(op1.sdiv(op2), 0, 0, 0)

    def test_both_1_hemisphere(self):
        # Both in 1-hemisphere
        op1 = StridedInterval(bits=4, stride=1, lower_bound=-6, upper_bound=-4)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=-3, upper_bound=-1)
        # sdiv: <4>0x1[0x1, 0x6]
        # udiv: <4>0x0[0x0, 0x0]
        assert check_si_fields(op1.sdiv(op2), 1, 1, 6)
        assert check_si_fields(op1.udiv(op2), 0, 0, 0)

    def test_one_0_one_1_hemisphere(self):
        # one in 0-hemisphere and one in 1-hemisphere
        op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=4)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=-5, upper_bound=-1)
        # sdiv: <4>0x1[0xc, 0xf]
        # udiv: <4>0x0[0x0, 0x0]
        assert check_si_fields(op1.sdiv(op2), 1, 12, 15)
        assert check_si_fields(op1.udiv(op2), 0, 0, 0)

    # Overlapping

    def test_overlapping_both_0_hemisphere(self):
        # case a of figure 2
        # Both in 0-hemisphere
        op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=3)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=0, upper_bound=6)
        # sdiv: <4>0x1[0x0, 0x3]
        # udiv: <4>0x1[0x0, 0x3]
        assert check_si_fields(op1.sdiv(op2), 1, 0, 3)
        assert check_si_fields(op1.udiv(op2), 1, 0, 3)

    def test_overlapping_both_1_hemisphere(self):
        # Both in 1-hemisphere
        op1 = StridedInterval(bits=4, stride=1, lower_bound=-6, upper_bound=-4)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=-7, upper_bound=-1)
        # sdiv: <4>0x1[0x0, 0x6]
        # udiv: <4>0x1[0x0, 0x1]
        assert check_si_fields(op1.sdiv(op2), 1, 0, 6)
        assert check_si_fields(op1.udiv(op2), 1, 0, 1)

    def test_overlapping_case_b(self):
        # case b Fig 2
        op1 = StridedInterval(bits=4, stride=1, lower_bound=3, upper_bound=-6)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=7, upper_bound=5)
        # sdiv: <4>0x1[0x0, 0xf]
        # udiv: <4>0x1[0x0, 0xa]
        assert check_si_fields(op1.sdiv(op2), 1, 0, 15)
        assert check_si_fields(op1.udiv(op2), 1, 0, 10)

    def test_overlapping_case_c(self):
        # case c Fig 2
        op1 = StridedInterval(bits=4, stride=1, lower_bound=3, upper_bound=-6)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=-2, upper_bound=5)
        # sdiv: <4>0x1[0x0, 0xe]
        # udiv: <4>0x1[0x0, 0xa]
        assert check_si_fields(op1.sdiv(op2), 1, 0, 14)
        assert check_si_fields(op1.udiv(op2), 1, 0, 10)

    # Strided Tests

    def test_strided_both_0_hemisphere(self):
        # Both in 0-hemisphere
        op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=3)
        op2 = StridedInterval(bits=4, stride=2, lower_bound=4, upper_bound=6)
        # sdiv: <4>0x0[0x0, 0x0]
        # udiv: <4>0x0[0x0, 0x0]
        assert check_si_fields(op1.sdiv(op2), 0, 0, 0)
        assert check_si_fields(op1.udiv(op2), 0, 0, 0)

    def test_strided_both_1_hemisphere(self):
        # Both in 1-hemisphere
        op1 = StridedInterval(bits=4, stride=1, lower_bound=-6, upper_bound=-4)
        op2 = StridedInterval(bits=4, stride=2, lower_bound=-3, upper_bound=-1)
        # sdiv: <4>0x1[0x1, 0x6]
        # udiv: <4>0x0[0x0, 0x0]
        assert check_si_fields(op1.sdiv(op2), 1, 1, 6)
        assert check_si_fields(op1.udiv(op2), 0, 0, 0)

    def test_strided_overlapping_case_1(self):
        # Overlapping case 1
        op1 = StridedInterval(bits=4, stride=2, lower_bound=3, upper_bound=-7)
        op2 = StridedInterval(bits=4, stride=3, lower_bound=7, upper_bound=3)
        # sdiv: <4>0x1[0x9, 0x7]
        # udiv: <4>0x1[0x0, 0x9]
        assert check_si_fields(op1.sdiv(op2), 1, 9, 7)
        assert check_si_fields(op1.udiv(op2), 1, 0, 9)

    def test_strided_overlapping_case_2(self):
        # Overlapping case 2
        op1 = StridedInterval(bits=4, stride=2, lower_bound=3, upper_bound=-7)
        op2 = StridedInterval(bits=4, stride=2, lower_bound=1, upper_bound=3)
        # sdiv: <4>0x1[0x1, 0xd]
        # udiv: <4>0x1[0x1, 0x9]
        assert check_si_fields(op1.sdiv(op2), 1, 1, 13)
        assert check_si_fields(op1.udiv(op2), 1, 1, 9)


class TestMultiplication(unittest.TestCase):
    # non-overlapping
    def test_simple_1(self):
        # simple case 1
        op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=3)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=2, upper_bound=2)
        # <4>0x2[0x2, 0x6]
        assert check_si_fields(op1.mul(op2), 2, 2, 6)

    def test_simple_2(self):
        # simple case 2
        op1 = StridedInterval(bits=4, stride=1, lower_bound=-6, upper_bound=-4)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=1)
        # <4>0x1[0xa, 0xc]
        assert check_si_fields(op1.mul(op2), 1, 10, 12)

    def test_simple_3(self):
        # simple case 3
        op1 = StridedInterval(bits=4, stride=1, lower_bound=3, upper_bound=-2)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=-2, upper_bound=-2)
        # Stride should be 2.
        # NOTE: previous result was: <4>0x1[0x4, 0x0] which is wrong.
        # possible values of 1[3,e] * 0[e,e] on 4 bits are [a, 8, 6, 4, 2, 0, e, c]
        # in the previous SI 2 was not present.
        assert check_si_fields(op1.mul(op2), 2, 2, 0)

    def test_simple_4(self):
        # simple case 4 : Result should be zero
        op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=3)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=0, upper_bound=0)
        # <4>0x0[0x0, 0x0]
        assert check_si_fields(op1.mul(op2), 0, 0, 0)

    def test_both_0_hemisphere(self):
        # simple case 4 : Result should be zero
        # Both in 0-hemisphere
        op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=3)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=4, upper_bound=6)
        # Result: <4>0x1[0x4, 0x2]
        assert check_si_fields(op1.mul(op2), 1, 4, 2)

    def test_both_1_hemisphere(self):
        # Both in 1-hemisphere
        op1 = StridedInterval(bits=4, stride=1, lower_bound=-6, upper_bound=-4)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=-3, upper_bound=-1)
        # Result <4>0x1[0x4, 0x2]
        assert check_si_fields(op1.mul(op2), 1, 4, 2)

    def test_one_0_one_1_hemisphere(self):
        # one in 0-hemisphere and one in 1-hemisphere
        op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=4)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=-5, upper_bound=-1)
        # TOP
        assert check_si_fields(op1.mul(op2), 1, 0, 15)

    # Overlapping

    def test_overlapping_both_0_hemisphere(self):
        # case a of figure 2
        # Both in 0-hemisphere
        op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=3)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=0, upper_bound=6)
        # TOP
        assert check_si_fields(op1.mul(op2), 1, 0, 15)

    def test_overlapping_both_1_hemisphere(self):
        # Both in 1-hemisphere
        op1 = StridedInterval(bits=4, stride=1, lower_bound=-6, upper_bound=-4)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=-7, upper_bound=-1)
        # TOP
        assert check_si_fields(op1.mul(op2), 1, 0, 15)

    def test_overlapping_case_b(self):
        # case b Fig 2
        op1 = StridedInterval(bits=4, stride=1, lower_bound=3, upper_bound=-6)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=7, upper_bound=5)
        # <4>0x1[0x0, 0xf]
        assert check_si_fields(op1.mul(op2), 1, 0, 15)

    def test_overlapping_case_c(self):
        # case c Fig 2
        op1 = StridedInterval(bits=4, stride=1, lower_bound=3, upper_bound=-6)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=-2, upper_bound=5)
        # <4>0x1[0x0, 0xf]
        assert check_si_fields(op1.mul(op2), 1, 0, 15)

    # Strided Tests

    def test_strided_both_0_hemisphere(self):
        # Both in 0-hemisphere
        op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=3)
        op2 = StridedInterval(bits=4, stride=2, lower_bound=4, upper_bound=6)
        # <4>0x1[0x4, 0x2]
        assert check_si_fields(op1.mul(op2), 1, 4, 2)

    def test_strided_both_1_hemisphere(self):
        # Both in 1-hemisphere
        op1 = StridedInterval(bits=4, stride=1, lower_bound=-6, upper_bound=-4)
        op2 = StridedInterval(bits=4, stride=2, lower_bound=-3, upper_bound=-1)
        # <4>0x1[0x4, 0x2]
        assert check_si_fields(op1.mul(op2), 1, 4, 2)

    def test_strided_overlapping_case_1(self):
        # Overlapping case 1
        op1 = StridedInterval(bits=4, stride=2, lower_bound=3, upper_bound=-7)
        op2 = StridedInterval(bits=4, stride=3, lower_bound=7, upper_bound=3)
        # <4>0x1[0x0, 0xf]
        assert check_si_fields(op1.mul(op2), 1, 0, 15)

    def test_strided_overlapping_case_2(self):
        # Overlapping case 2
        op1 = StridedInterval(bits=4, stride=2, lower_bound=3, upper_bound=-7)
        op2 = StridedInterval(bits=4, stride=2, lower_bound=1, upper_bound=3)
        # TOP
        assert check_si_fields(op1.mul(op2), 1, 0, 15)


class TestSubtraction(unittest.TestCase):
    def test_basic_interval_1(self):
        # Basic Interval Tests
        op1 = StridedInterval(bits=4, stride=1, lower_bound=-2, upper_bound=7)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=0, upper_bound=-6)
        # Result should be TOP
        assert check_si_fields(op1.sub(op2), 1, 0, 15)

    def test_basic_interval_2(self):
        op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=7)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=2, upper_bound=6)
        # Result should be 1,[-5, 5]
        assert check_si_fields(op1.sub(op2), 1, -5, 5)

    def test_basic_interval_3(self):
        op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=7)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=2, upper_bound=2)
        # Result should be 1,[15, 5]
        assert check_si_fields(op1.sub(op2), 1, 15, 5)

    def test_basic_interval_4(self):
        op1 = StridedInterval(bits=4, stride=1, lower_bound=-2, upper_bound=7)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=2, upper_bound=2)
        # Result should be 1,[-4, 5]
        assert check_si_fields(op1.sub(op2), 1, -4, 5)

    def test_basic_interval_5(self):
        op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=7)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=0, upper_bound=0)
        # Result should be 1,[1, 7]
        assert check_si_fields(op1.sub(op2), 1, 1, 7)

    def test_basic_interval_6(self):
        op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=7)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=-5, upper_bound=-1)
        # Result should be 1,[2, 12]
        assert check_si_fields(op1.sub(op2), 1, 2, 12)

    # Strided Tests
    def test_strided_1(self):
        op1 = StridedInterval(bits=4, stride=2, lower_bound=-2, upper_bound=7)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=0, upper_bound=-6)
        # Result should be TOP
        assert check_si_fields(op1.sub(op2), 1, 0, 15)

    def test_strided_2(self):
        op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=7)
        op2 = StridedInterval(bits=4, stride=2, lower_bound=2, upper_bound=6)
        # Result should be 1,[11, 5]
        assert check_si_fields(op1.sub(op2), 1, 11, 5)


class TestAddition(unittest.TestCase):
    def test_interval_1(self):
        # Basic Interval Tests
        op1 = StridedInterval(bits=4, stride=1, lower_bound=-2, upper_bound=7)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=0, upper_bound=-6)
        # Result should be TOP
        assert check_si_fields(op1.add(op2), 1, 0, 15)

    def test_interval_2(self):
        op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=7)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=2, upper_bound=6)
        # Result should be 1,[3, 13]
        assert check_si_fields(op1.add(op2), 1, 3, 13)

    def test_interval_3(self):
        op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=7)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=2, upper_bound=2)
        # Result should be 1,[3, 9]
        assert check_si_fields(op1.add(op2), 1, 3, 9)

    def test_interval_4(self):
        op1 = StridedInterval(bits=4, stride=1, lower_bound=-2, upper_bound=7)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=2, upper_bound=2)
        # Result should be 1,[0,9]
        assert check_si_fields(op1.add(op2), 1, 0, 9)

    def test_interval_5(self):
        op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=7)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=0, upper_bound=0)
        # Result should be 1,[1,7]
        assert check_si_fields(op1.add(op2), 1, 1, 7)

    def test_interval_6(self):
        op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=7)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=-5, upper_bound=-1)
        # Result should be 1,[-4, 6]
        assert check_si_fields(op1.add(op2), 1, -4, 6)

    # Strided Tests
    def test_strided_1(self):
        op1 = StridedInterval(bits=4, stride=2, lower_bound=-2, upper_bound=7)
        op2 = StridedInterval(bits=4, stride=1, lower_bound=0, upper_bound=-6)
        # Result should be TOP
        assert check_si_fields(op1.add(op2), 1, 0, 15)

    def test_strided_2(self):
        op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=7)
        op2 = StridedInterval(bits=4, stride=2, lower_bound=2, upper_bound=6)
        # Result should be 1,[3, 13]
        assert check_si_fields(op1.add(op2), 1, 3, 13)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import logging
from contextlib import suppress
from typing import TYPE_CHECKING

import claripy
from claripy.errors import BackendError

if TYPE_CHECKING:
    from claripy.ast import Bool

log = logging.getLogger(__name__)


def is_true(expr: Bool) -> bool:
    """Checks if a boolean expression is trivially True.

    A false result does not necessarily mean that the expression is False, but
    rather that it is not trivially True.
    """

    with suppress(BackendError):
        return claripy.backends.concrete.is_true(expr)

    log.debug("Unable to tell the truth-value of this expression")
    return False


def is_false(expr: Bool) -> bool:
    """Checks if a boolean expression is trivially False.

    A false result does not necessarily mean that the expression is False, but
    rather that it is not trivially False.
    """

    with suppress(BackendError):
        return claripy.backends.concrete.is_false(expr)

    log.debug("Unable to tell the truth-value of this expression")
    return False



from __future__ import annotations

from typing import TypeVar, cast
from weakref import WeakValueDictionary

import claripy
from claripy.ast import Base

T = TypeVar("T", bound=Base)

#
# This code handles burrowing ITEs deeper into the ast and excavating
# them to shallower levels.
#

burrowed_cache: WeakValueDictionary[int, Base] = WeakValueDictionary()
excavated_cache: WeakValueDictionary[int, Base] = WeakValueDictionary()


def _burrow_ite(expr: T) -> T:
    if expr.op != "If":
        return expr.make_like(expr.op, [(burrow_ite(a) if isinstance(a, Base) else a) for a in expr.args])

    if not all(isinstance(a, Base) for a in expr.args):
        return expr

    old_true = expr.args[1]
    old_false = expr.args[2]

    if old_true.op != old_false.op or len(old_true.args) != len(old_false.args):
        return expr

    if old_true.op == "If":
        # let's no go into this right now
        return expr

    if any(a.is_leaf() for a in expr.args):
        # burrowing through these is pretty funny
        return expr

    matches = [old_true.args[i] is old_false.args[i] for i in range(len(old_true.args))]
    if matches.count(True) != 1 or all(matches):
        # TODO: handle multiple differences for multi-arg ast nodes
        # print("wrong number of matches:",matches,old_true,old_false)
        return expr

    different_idx = matches.index(False)
    inner_if = claripy.If(expr.args[0], old_true.args[different_idx], old_false.args[different_idx])
    new_args = list(old_true.args)
    new_args[different_idx] = burrow_ite(inner_if)
    return old_true.__class__(old_true.op, new_args, length=expr.length)


def _excavate_ite(expr: T) -> T:
    ast_queue = [iter([expr])]
    arg_queue = []
    op_queue = []

    while ast_queue:
        try:
            ast = next(ast_queue[-1])

            if not isinstance(ast, Base):
                arg_queue.append(ast)
                continue

            if ast.is_leaf():
                arg_queue.append(ast)
                continue

            if ast.annotations:
                arg_queue.append(ast)
                continue

            op_queue.append(ast)
            ast_queue.append(iter(ast.args))

        except StopIteration:
            ast_queue.pop()

            if op_queue:
                op = op_queue.pop()

                args = arg_queue[-len(op.args) :]
                del arg_queue[-len(op.args) :]

                ite_args = [isinstance(a, Base) and a.op == "If" for a in args]

                if op.op == "If":
                    # if we are an If, call the If handler so that we can take advantage of its simplifiers
                    excavated = claripy.If(*args)

                elif ite_args.count(True) == 0:
                    # if there are no ifs that came to the surface, there's nothing more to do
                    excavated = op.make_like(op.op, args, simplify=True)

                else:
                    # this gets called when we're *not* in an If, but there are Ifs in the args.
                    # it pulls those Ifs out to the surface.
                    cond = args[ite_args.index(True)].args[0]
                    new_true_args = []
                    new_false_args = []

                    for a in args:
                        if not isinstance(a, Base) or a.op != "If":
                            new_true_args.append(a)
                            new_false_args.append(a)
                        elif a.args[0] is cond:
                            new_true_args.append(a.args[1])
                            new_false_args.append(a.args[2])
                        elif a.args[0] is ~cond:
                            new_true_args.append(a.args[2])
                            new_false_args.append(a.args[1])
                        else:
                            # weird conditions -- giving up!
                            excavated = op.make_like(op.op, args, simplify=True)
                            break

                    else:
                        excavated = claripy.If(
                            cond,
                            op.make_like(op.op, new_true_args, simplify=True),
                            op.make_like(op.op, new_false_args, simplify=True),
                        )

                # continue
                arg_queue.append(excavated)

    assert len(op_queue) == 0, "op_queue is not empty"
    assert len(ast_queue) == 0, "ast_queue is not empty"
    assert len(arg_queue) == 1, ("arg_queue has unexpected length", len(arg_queue))

    return arg_queue.pop()


def burrow_ite(expr: T) -> T:
    """
    Returns an equivalent AST that "burrows" the ITE expressions as deep as
    possible into the ast, for simpler printing.
    """
    if expr.hash() in burrowed_cache and burrowed_cache[expr.hash()] is not None:
        return cast("T", burrowed_cache[expr.hash()])

    burrowed = _burrow_ite(expr)
    burrowed_cache[burrowed.hash()] = burrowed
    burrowed_cache[expr.hash()] = burrowed
    return burrowed


def excavate_ite(expr: T) -> T:
    """
    Returns an equivalent AST that "excavates" the ITE expressions out as far as
    possible toward the root of the AST, for processing in static analyses.
    """
    if expr.hash() in excavated_cache and excavated_cache[expr.hash()] is not None:
        return cast("T", excavated_cache[expr.hash()])

    excavated = _excavate_ite(expr)
    excavated_cache[excavated.hash()] = excavated
    excavated_cache[expr.hash()] = excavated
    return excavated



# pylint: disable=missing-class-docstring,no-self-use
from __future__ import annotations

import unittest

import claripy


class TestConcreteBackend(unittest.TestCase):
    def test_concrete(self):
        a = claripy.BVV(10, 32)
        b = claripy.BoolV(True)

        assert isinstance(claripy.backends.concrete.convert(a), claripy.backends.backend_concrete.BVV)
        assert isinstance(claripy.backends.concrete.convert(b), bool)

        a = claripy.BVV(1337, 32)
        b = a[31:16]
        c = claripy.BVV(0, 16)
        assert b is c

        bc = claripy.backends.concrete
        d = claripy.BVV(-1, 32)
        assert bc.convert(d) == 0xFFFFFFFF

        e = claripy.BVV(2**32 + 1337, 32)
        assert bc.convert(e) == 1337

    def test_concrete_fp(self):
        f = claripy.FPV(1.0, claripy.FSORT_FLOAT)
        assert claripy.backends.concrete.eval(f, 2) == (1.0,)


if __name__ == "__main__":
    unittest.main()



# pylint: disable=missing-class-docstring,no-self-use
from __future__ import annotations

import unittest

import claripy

KEEP_TEST_PERFORMANT = True


class TestStrings(unittest.TestCase):
    def get_solver(self):
        return claripy.SolverStrings(backend=claripy.backends.z3)

    def test_concat(self):
        str_concrete = claripy.StringV("conc")
        str_symbol = claripy.StringS("symb_concat", explicit_name=True)
        solver = self.get_solver()
        res = str_concrete + str_symbol
        solver.add(res == claripy.StringV("concrete"))
        self.assertTrue(solver.satisfiable())
        result = solver.eval(str_symbol, 2)
        self.assertEqual(1, len(result))
        self.assertEqual("rete", result[0])

        result = solver.eval_to_ast(str_symbol, 2)
        self.assertEqual([claripy.StringV("rete")], list(result))

    def test_concat_simplification(self):
        solver = self.get_solver()
        str_concrete = claripy.StringV("conc")
        res = str_concrete + str_concrete + str_concrete
        res2 = claripy.StrConcat(str_concrete, str_concrete)
        res3 = claripy.StrConcat(res2, str_concrete)
        solver.add(res == res3)
        self.assertTrue(solver.satisfiable())
        result = solver.eval(str_concrete, 2)
        self.assertEqual(["conc"], list(result))

    def test_substr(self):
        str_symbol = claripy.StringS("symb_subst", explicit_name=True)
        solver = self.get_solver()
        solver.add(claripy.StrSubstr(1, 2, str_symbol) == claripy.StringV("o"))
        self.assertTrue(solver.satisfiable())
        results = solver.eval(str_symbol, 2 if KEEP_TEST_PERFORMANT else 100)
        self.assertEqual(len(results), 2 if KEEP_TEST_PERFORMANT else 100)
        for s in results:
            self.assertTrue(s[1:2] == "o")

    def test_substr_simplification(self):
        str_concrete = claripy.StringV("concrete")
        solver = self.get_solver()
        # TODO: Make sure that semantics of Substr match the ones of SMTLib substr
        solver.add(claripy.StrSubstr(1, 2, str_concrete) == claripy.StringV("on"))
        self.assertTrue(solver.satisfiable())
        result = solver.eval(str_concrete, 2)
        self.assertEqual(list(result), ["concrete"])

    def test_replace(self):
        str_to_replace_symb = claripy.StringS("symb_repl", explicit_name=True)
        sub_str_to_repl = claripy.StringV("a")
        replacement = claripy.StringV("b")
        solver = self.get_solver()
        repl_stringa = claripy.StrReplace(str_to_replace_symb, sub_str_to_repl, replacement)
        solver.add(repl_stringa == claripy.StringV("cbne"))
        self.assertTrue(solver.satisfiable())

        result = solver.eval(repl_stringa, 2)
        self.assertEqual(list(result), ["cbne"])

        result = solver.eval(str_to_replace_symb, 2 if KEEP_TEST_PERFORMANT else 100)
        self.assertEqual(set(result), {"cbne", "cane"})

    def test_replace_simplification(self):
        str_to_replace = claripy.StringV("cane")
        sub_str_to_repl = claripy.StringV("a")
        replacement = claripy.StringV("b")
        repl_stringa = claripy.StrReplace(str_to_replace, sub_str_to_repl, replacement)
        solver = self.get_solver()
        solver.add(repl_stringa == claripy.StringV("cbne"))

        self.assertTrue(solver.satisfiable())

        result = solver.eval(repl_stringa, 2)
        self.assertEqual(["cbne"], list(result))

        result = solver.eval(str_to_replace, 2)
        self.assertEqual(["cane"], list(result))

    def test_ne(self):
        str_symb = claripy.StringS("symb_ne", explicit_name=True)
        solver = self.get_solver()
        solver.add(str_symb != claripy.StringV("concrete"))
        self.assertTrue(solver.satisfiable())

        result = solver.eval(str_symb, 4 if KEEP_TEST_PERFORMANT else 100)
        self.assertTrue("concrete" not in result)

    def test_length(self):
        str_symb = claripy.StringS("symb_length", explicit_name=True)
        solver = self.get_solver()
        # TODO: How do we want to deal with the size of a symbolic string?
        solver.add(claripy.StrLen(str_symb) == 14)
        self.assertTrue(solver.satisfiable())

        result = solver.eval(str_symb, 4 if KEEP_TEST_PERFORMANT else 100)
        for r in result:
            self.assertTrue(len(r) == 14)

    def test_length_simplification(self):
        str_concrete = claripy.StringV("concrete")
        solver = self.get_solver()
        solver.add(claripy.StrLen(str_concrete) == 8)
        self.assertTrue(solver.satisfiable())

        result = solver.eval(str_concrete, 2)
        self.assertEqual(["concrete"], list(result))
        for r in result:
            self.assertTrue(len(r) == 8)

    def test_or(self):
        str_symb = claripy.StringS("Symb_or", explicit_name=True)
        solver = self.get_solver()
        res = claripy.Or((str_symb == claripy.StringV("abc")), (str_symb == claripy.StringV("ciao")))
        solver.add(res)
        self.assertTrue(solver.satisfiable())

        result = solver.eval(str_symb, 3 if KEEP_TEST_PERFORMANT else 100)
        self.assertEqual({"ciao", "abc"}, set(result))

    def test_lt_etc(self):
        str_symb = claripy.StringS("Symb_2")
        solver = self.get_solver()
        c1 = claripy.StrLen(str_symb) <= 4
        c2 = claripy.StrLen(str_symb) < 4
        c3 = claripy.StrLen(str_symb) >= 4
        c4 = claripy.StrLen(str_symb) > 4
        solver.add(c1)
        solver.add(c2)
        solver.add(c3)
        solver.add(c4)
        self.assertFalse(solver.satisfiable())

    def test_substr_BV_concrete_index(self):
        str_symbol = claripy.StringS("symb_subst", explicit_name=True)
        solver = self.get_solver()
        bv1 = claripy.BVV(1, 32)
        bv2 = claripy.BVV(2, 32)
        res = claripy.StrSubstr(bv1, bv2, str_symbol) == claripy.StringV("on")
        solver.add(res)
        self.assertTrue(solver.satisfiable())
        self.assertEqual("on", solver.eval(str_symbol, 1)[0][1:3])

    def test_substr_BV_symbolic_index(self):
        str_symbol = claripy.StringS("symb_subst", explicit_name=True)
        solver = self.get_solver()
        start = claripy.BVS("start_idx", 32)
        count = claripy.BVS("count", 32)
        res = claripy.StrSubstr(start, count, str_symbol) == claripy.StringV("on")
        solver.add(res)
        self.assertTrue(solver.satisfiable())
        self.assertEqual("on", solver.eval(str_symbol, 1, extra_constraints=(start == 0, count == 2))[0][0:2])
        self.assertEqual("on", solver.eval(str_symbol, 1, extra_constraints=(start == 1, count == 2))[0][1:3])
        self.assertEqual("on", solver.eval(str_symbol, 1, extra_constraints=(start == 2, count == 2))[0][2:4])

        self.assertEqual("on", solver.eval(str_symbol, 1, extra_constraints=(start == 2, count == 3))[0][2:4])
        self.assertEqual("on", solver.eval(str_symbol, 1, extra_constraints=(start == 2, count == 4))[0][2:4])

        self.assertEqual("on", solver.eval(str_symbol, 1, extra_constraints=(start == 0, count == 3))[0])
        self.assertEqual("on", solver.eval(str_symbol, 1, extra_constraints=(start == 1, count == 4))[0][1:])

    def test_substr_BV_mixed_index(self):
        str_symbol = claripy.StringS("symb_subst", explicit_name=True)
        solver = self.get_solver()
        start = claripy.BVS("symb_subst_start_idx", 32, explicit_name=True)
        count = claripy.BVV(2, 32)
        res = claripy.StrSubstr(start, count, str_symbol) == claripy.StringV("on")
        solver.add(res)
        self.assertTrue(solver.satisfiable())
        self.assertEqual("on", solver.eval(str_symbol, 1, extra_constraints=(start == 0,))[0][0:2])
        self.assertEqual("on", solver.eval(str_symbol, 1, extra_constraints=(start == 1,))[0][1:3])
        self.assertEqual("on", solver.eval(str_symbol, 1, extra_constraints=(start == 2,))[0][2:4])

    def test_contains(self):
        str_symb = claripy.StringS("symb_contains", explicit_name=True)
        res = claripy.StrContains(str_symb, claripy.StringV("an"))
        solver = self.get_solver()
        solver.add(res)
        self.assertTrue(solver.satisfiable())
        solutions = solver.eval(str_symb, 4 if KEEP_TEST_PERFORMANT else 100)
        for sol in solutions:
            self.assertTrue("an" in sol)

    def test_contains_simplification(self):
        str_concrete = claripy.StringV("concrete")
        solver = self.get_solver()
        res = claripy.StrContains(str_concrete, claripy.StringV("nc"))
        solver.add(res)
        self.assertTrue(solver.satisfiable())
        self.assertEqual((), tuple(solver.constraints))
        self.assertEqual(("concrete",), solver.eval(str_concrete, 2))
        self.assertEqual((True,), solver.eval(res, 2))

    def test_prefix(self):
        str_symb = claripy.StringS("symb_prefix", explicit_name=True)
        res = claripy.StrPrefixOf(claripy.StringV("an"), str_symb)
        solver = self.get_solver()
        solver.add(res)
        self.assertTrue(solver.satisfiable())

        solutions = solver.eval(str_symb, 4 if KEEP_TEST_PERFORMANT else 100)
        for sol in solutions:
            self.assertTrue(sol.startswith("an"))

    def test_suffix(self):
        str_symb = claripy.StringS("symb_suffix", explicit_name=True)
        res = claripy.StrSuffixOf(claripy.StringV("an"), str_symb)
        solver = self.get_solver()
        solver.add(res)
        self.assertTrue(solver.satisfiable())

        solutions = solver.eval(str_symb, 4 if KEEP_TEST_PERFORMANT else 100)
        for sol in solutions:
            self.assertTrue(sol.endswith("an"))

    def test_prefix_simplification(self):
        str_concrete = claripy.StringV("concrete")
        solver = self.get_solver()
        res = claripy.StrPrefixOf(claripy.StringV("conc"), str_concrete)
        solver.add(res)
        self.assertTrue(solver.satisfiable())
        self.assertEqual((), tuple(solver.constraints))
        self.assertEqual(("concrete",), solver.eval(str_concrete, 2))
        self.assertEqual((True,), solver.eval(res, 2))

    def test_suffix_simplification(self):
        str_concrete = claripy.StringV("concrete")
        solver = self.get_solver()
        res = claripy.StrSuffixOf(claripy.StringV("rete"), str_concrete)
        solver.add(res)
        self.assertTrue(solver.satisfiable())
        self.assertEqual((), tuple(solver.constraints))
        self.assertEqual(("concrete",), solver.eval(str_concrete, 2))
        self.assertEqual((True,), solver.eval(res, 2))

    def test_index_of(self):
        str_symb = claripy.StringS("symb_suffix", explicit_name=True)
        res = claripy.StrIndexOf(str_symb, claripy.StringV("an"), 0)
        solver = self.get_solver()

        target_idx = 4 if KEEP_TEST_PERFORMANT else 100
        solver.add(res == target_idx)
        self.assertTrue(solver.satisfiable())

        solutions = solver.eval(str_symb, 4 if KEEP_TEST_PERFORMANT else 100)
        for sol in solutions:
            self.assertEqual("an", sol[target_idx : target_idx + 2])

        self.assertEqual((target_idx,), solver.eval(res, 2))

    def test_index_of_simplification(self):
        str_concrete = claripy.StringV("concrete")
        solver = self.get_solver()
        res = claripy.StrIndexOf(str_concrete, claripy.StringV("rete"), 0)
        target_idx = 4 if KEEP_TEST_PERFORMANT else 100
        solver.add(res == target_idx)
        self.assertTrue(solver.satisfiable())
        self.assertEqual((), tuple(solver.constraints))
        self.assertEqual((target_idx,), solver.eval(res, 2))

    @unittest.skip("Usually hangs")
    def test_index_of_symbolic_start_idx(self):
        str_symb = claripy.StringS("symb_index_of", explicit_name=True)
        start_idx = claripy.BVS("symb_start_idx", 32, explicit_name=True)

        solver = self.get_solver()

        solver.add(start_idx > 32)
        solver.add(start_idx < 35)
        res = claripy.StrIndexOf(str_symb, claripy.StringV("an"), start_idx)

        solver.add(res != -1)
        solver.add(res < 38)
        self.assertTrue(solver.satisfiable())
        self.assertEqual({33, 34, 35, 36, 37}, set(solver.eval(res, 10)))

        strs = solver.eval(str_symb, 10 if KEEP_TEST_PERFORMANT else 100)
        for s in strs:
            self.assertTrue(32 < s.index("an") < 38)

    def test_str_to_int(self):
        str_symb = claripy.StringS("symb_strtoint", explicit_name=True)
        res = claripy.StrToInt(str_symb)
        solver = self.get_solver()
        target_num = 12 if KEEP_TEST_PERFORMANT else 100000
        solver.add(res == target_num)
        self.assertTrue(solver.satisfiable())

        solutions = solver.eval(str_symb, 2 if KEEP_TEST_PERFORMANT else 1000000)
        for sol in solutions:
            self.assertTrue(int(sol) == target_num)

    def test_str_to_int_simplification(self):
        target_num = 12 if not KEEP_TEST_PERFORMANT else 1000000

        str_concrete = claripy.StringV(str(target_num))
        solver = self.get_solver()
        res = claripy.StrToInt(str_concrete)

        solver.add(res == target_num)
        self.assertTrue(solver.satisfiable())
        self.assertEqual((), tuple(solver.constraints))
        self.assertEqual((target_num,), solver.eval(res, 2))



from __future__ import annotations

from claripy.errors import ClaripyError


class ClaripyVSAError(ClaripyError):
    """ClaripyVSAError is the base class for all errors raised by claripy's
    VSA backend.
    """


class ClaripyVSAOperationError(ClaripyVSAError):
    """ClaripyVSAOperationError is raised when an operation is attempted on VSA
    objects that is not supported.
    """



from __future__ import annotations

import logging

from claripy import backends
from claripy.errors import BackendError, ClaripyFrontendError

from .constrained_frontend import ConstrainedFrontend

log = logging.getLogger(__name__)


class LightFrontend(ConstrainedFrontend):
    """LightFrontend is an extremely simple frontend that is used primarily for
    the purpose of quickly evaluating expressions using the concrete and VSA
    backends.
    """

    def __init__(self, solver_backend, **kwargs):
        # since the LightFrontend mostly cannot handle extra_constriants, it horribly misuses the cache.
        # Because of this, we have to disable the caching here.
        super().__init__(**kwargs)
        self._solver_backend = solver_backend

    def _blank_copy(self, c):
        super()._blank_copy(c)
        c._solver_backend = self._solver_backend

    def _copy(self, c):
        super()._copy(c)
        c._solver_backend = self._solver_backend

    #
    # Serialization support
    #

    def __getstate__(self):
        return self._solver_backend.__class__.__name__, super().__getstate__()

    def __setstate__(self, s):
        solver_backend_name, base_state = s
        super().__setstate__(base_state)
        self._solver_backend = backends.backends_by_type[solver_backend_name]

    #
    # Light functionality
    #

    def eval(self, e, n, extra_constraints=(), exact=None):
        try:
            return tuple(self._solver_backend.eval(e, n))
        except BackendError as err:
            raise ClaripyFrontendError("Light solver can't handle this eval().") from err

    def batch_eval(self, exprs, n, extra_constraints=(), exact=None):
        try:
            return self._solver_backend._batch_eval(exprs, n)
        except BackendError as err:
            raise ClaripyFrontendError("Light solver can't handle this batch_eval().") from err

    def max(self, e, extra_constraints=(), signed=False, exact=None):
        try:
            return self._solver_backend.max(e, signed=signed)
        except BackendError as err:
            raise ClaripyFrontendError("Light solver can't handle this max().") from err

    def min(self, e, extra_constraints=(), signed=False, exact=None):
        try:
            return self._solver_backend.min(e, signed=signed)
        except BackendError as err:
            raise ClaripyFrontendError("Light solver can't handle this min().") from err

    def solution(self, e, v, extra_constraints=(), exact=None):
        try:
            return self._solver_backend.solution(e, v)
        except BackendError as err:
            raise ClaripyFrontendError("Light solver can't handle this solution().") from err

    def is_true(self, e, extra_constraints=(), exact=None):
        try:
            return self._solver_backend.is_true(e)
        except BackendError:
            log.info("Light solver can't handle this is_true().")
            return False

    def is_false(self, e, extra_constraints=(), exact=None):
        try:
            return self._solver_backend.is_false(e)
        except BackendError:
            log.info("Light solver can't handle this is_false().")
            return False

    def satisfiable(self, extra_constraints=(), exact=None):
        return not any(
            self.is_false(c, extra_constraints=extra_constraints, exact=exact)
            for c in reversed(self.constraints + list(extra_constraints))
        )

    #
    # Merging and splitting
    #

    def merge(self, others, merge_conditions, common_ancestor=None):
        return (
            self._solver_backend.__class__.__name__ == "BackendZ3",
            ConstrainedFrontend.merge(self, others, merge_conditions, common_ancestor=common_ancestor)[1],
        )



from __future__ import annotations


class ClaripyError(Exception):
    pass


class UnsatError(ClaripyError):
    pass


class ClaripyFrontendError(ClaripyError):
    pass


class ClaripySerializationError(ClaripyError):
    pass


class BackendError(ClaripyError):
    pass


class BackendUnsupportedError(BackendError):
    pass


class ClaripyZ3Error(ClaripyError):
    pass


class ClaripyBackendVSAError(BackendError):
    pass


class MissingSolverError(ClaripyError):
    pass


class ClaripySolverInterruptError(ClaripyError):
    pass


#
# AST errors
#


class ClaripyASTError(ClaripyError):
    pass


class ClaripyBalancerError(ClaripyASTError):
    pass


class ClaripyBalancerUnsatError(ClaripyBalancerError):
    pass


class ClaripyTypeError(ClaripyASTError):
    pass


class ClaripyValueError(ClaripyASTError):
    pass


class ClaripySizeError(ClaripyASTError):
    pass


class ClaripyOperationError(ClaripyASTError):
    pass


class ClaripyReplacementError(ClaripyASTError):
    pass


class ClaripyRecursionError(ClaripyOperationError):
    pass


class ClaripyZeroDivisionError(ClaripyOperationError, ZeroDivisionError):
    pass
